@@ -23,9 +23,14 @@
 #include <stdint.h>
 
 #include <libplacebo/common.h>
+#include <libplacebo/log.h>
 
 PL_API_BEGIN
 
+// These are not memory managed, and should represent compile-time constants
+typedef const char *pl_debug_tag;
+#define PL_DEBUG_TAG (__FILE__ ":" PL_TOSTRING(__LINE__))
+
 // Type of a shader input descriptor.
 enum pl_desc_type {
     PL_DESC_INVALID = 0,
@@ -115,17 +120,18 @@ struct pl_gpu_limits {
     // optimal alignment. For best performance, the corresponding field
     // should be aligned to a multiple of these. They will always be a power
     // of two.
-    uint32_t align_tex_xfer_stride; // optimal `pl_tex_transfer_params.stride_w/h`
+    size_t align_tex_xfer_pitch;    // optimal `pl_tex_transfer_params.row_pitch`
     size_t align_tex_xfer_offset;   // optimal `pl_tex_transfer_params.buf_offset`
 
     // --- pl_pass
-    size_t max_variables;       // maximum `pl_pass_params.num_variables`
+    size_t max_variable_comps;  // maximum components passed in variables
     size_t max_constants;       // maximum `pl_pass_params.num_constants`
     size_t max_pushc_size;      // maximum `push_constants_size`
+    size_t align_vertex_stride; // alignment of `pl_pass_params.vertex_stride`
     uint32_t max_dispatch[3];   // maximum dispatch size per dimension
 
-    // Note: At least one of `max_variables` or `max_ubo_size` is guaranteed to
-    // be nonzero.
+    // Note: At least one of `max_variable_comps` or `max_ubo_size` is
+    // guaranteed to be nonzero.
 
     // As a performance hint, the GPU may signal the number of command queues
     // it has for fragment and compute shaders, respectively. Users may use
@@ -141,10 +147,12 @@ struct pl_gpu_limits {
     uint32_t subgroup_size PL_DEPRECATED;
     int16_t min_gather_offset PL_DEPRECATED;
     int16_t max_gather_offset PL_DEPRECATED;
+    size_t max_variables PL_DEPRECATED; // see `max_variable_comps`
 };
 
-// Backwards compatibility alias
+// Backwards compatibility aliases
 #define max_xfer_size max_buf_size
+#define align_tex_xfer_stride align_tex_xfer_pitch
 
 // Some `pl_gpu` operations allow sharing GPU resources with external APIs -
 // examples include interop with other graphics APIs such as CUDA, and also
@@ -189,6 +197,9 @@ struct pl_shared_mem {
     size_t size;   // the total size of the memory referenced by this handle
     size_t offset; // the offset of the object within the referenced memory
 
+    // Note: `size` is optional for some APIs and handle types, in particular
+    // when importing DMABUFs or D3D11 textures.
+
     // For PL_HANDLE_DMA_BUF, this specifies the DRM format modifier that
     // describes this resource. Note that when importing `pl_buf`, this must
     // be DRM_FORMAT_MOD_LINEAR. For importing `pl_tex`, it can be any
@@ -215,7 +226,7 @@ typedef uint64_t pl_gpu_caps;
 enum PL_DEPRECATED {
     PL_GPU_CAP_COMPUTE          = 1 << 0, // see `pl_glsl_version.compute`
     PL_GPU_CAP_PARALLEL_COMPUTE = 1 << 1, // see `pl_gpu_limits.compute_queues`
-    PL_GPU_CAP_INPUT_VARIABLES  = 1 << 2, // see `pl_gpu_limits.max_variables`
+    PL_GPU_CAP_INPUT_VARIABLES  = 1 << 2, // see `pl_gpu_limits.max_variable_comps`
     PL_GPU_CAP_MAPPED_BUFFERS   = 1 << 3, // see `pl_gpu_limits.max_mapped_size`
     PL_GPU_CAP_BLITTABLE_1D_3D  = 1 << 4, // see `pl_gpu_limits.blittable_1d_3d`
     PL_GPU_CAP_SUBGROUPS        = 1 << 5, // see `pl_glsl_version.subgroup_size`
@@ -256,9 +267,12 @@ typedef const PL_STRUCT(pl_gpu) {
     pl_gpu_caps caps PL_DEPRECATED; // replaced by `glsl` and `limits`
 } *pl_gpu;
 
-// Helper function to align the given dimension (e.g. width or height) to a
-// multiple of the optimal texture transfer stride.
-int pl_optimal_transfer_stride(pl_gpu gpu, int dimension);
+// No longer functional. See `pl_gpu_limits.align_tex_xfer_pitch`.
+PL_DEPRECATED static inline int pl_optimal_transfer_stride(pl_gpu _gpu, int dim)
+{
+    (void) _gpu;
+    return dim;
+}
 
 enum pl_fmt_type {
     PL_FMT_UNKNOWN = 0, // also used for inconsistent multi-component formats
@@ -294,6 +308,7 @@ enum pl_fmt_caps {
 // Structure describing a texel/vertex format.
 PL_STRUCT(pl_fmt) {
     const char *name;       // symbolic name for this format (e.g. rgba32f)
+    uint64_t signature;     // unique but stable signature (for pass reusability)
 
     enum pl_fmt_type type;  // the format's data type and interpretation
     enum pl_fmt_caps caps;  // the features supported by this format
@@ -320,6 +335,7 @@ PL_STRUCT(pl_fmt) {
     bool opaque;
     bool emulated;
     size_t texel_size;      // total size in bytes per texel
+    size_t texel_align;     // texel alignment requirements (bytes)
     int host_bits[4];       // number of meaningful bits in host memory
     int sample_order[4];    // sampled index for each component, e.g.
                             // {2, 1, 0, 3} for BGRA textures
@@ -358,6 +374,9 @@ bool pl_fmt_is_ordered(pl_fmt fmt);
 // Returns whether or not a pl_fmt is sampled as a float (e.g. UNORM)
 bool pl_fmt_is_float(pl_fmt fmt);
 
+// Returns whether or not a pl_fmt supports a given DRM modifier.
+bool pl_fmt_has_modifier(pl_fmt fmt, uint64_t modifier);
+
 // Helper function to find a format with a given number of components and
 // minimum effective precision per component. If `host_bits` is set, then the
 // format will always be non-opaque, unpadded, ordered and have exactly this
@@ -412,18 +431,6 @@ void pl_timer_destroy(pl_gpu gpu, pl_timer *);
 // all of its results, or older results will be overwritten by newer ones.
 uint64_t pl_timer_query(pl_gpu gpu, pl_timer);
 
-// (Deprecated) Buffer usage type. This defines what types of operations may be
-// performed on a buffer. They are defined merely for backwards compatibility,
-// and correspond to merely enabling the respective usage flags.
-enum pl_buf_type {
-    PL_BUF_TEX_TRANSFER,  // no extra usage flags
-    PL_BUF_UNIFORM,       // enables `uniform`
-    PL_BUF_STORAGE,       // enables `storable`
-    PL_BUF_TEXEL_UNIFORM, // equivalent to PL_BUF_UNIFORM (when `format` is set)
-    PL_BUF_TEXEL_STORAGE, // equivalent to PL_BUF_STORAGE (when `format` is set)
-    PL_BUF_TYPE_COUNT,
-};
-
 enum pl_buf_mem_type {
     PL_BUF_MEM_AUTO = 0, // use whatever seems most appropriate
     PL_BUF_MEM_HOST,     // try allocating from host memory (RAM)
@@ -489,12 +496,15 @@ struct pl_buf_params {
     // Arbitrary user data. libplacebo does not use this at all.
     void *user_data;
 
-    // Deprecated. Setting a type now effectively just enables some of the
-    // buffer usage flags. See `pl_buf_type`. This field will be removed
-    // in the future.
-    enum pl_buf_type type PL_DEPRECATED;
+    // Arbitrary identifying tag. Used only for debugging purposes.
+    pl_debug_tag debug_tag;
 };
 
+#define pl_buf_params(...) (&(struct pl_buf_params) {   \
+        .debug_tag = PL_DEBUG_TAG,                      \
+        __VA_ARGS__                                     \
+    })
+
 // A generic buffer, which can be used for multiple purposes (texture transfer,
 // storage buffer, uniform buffer, etc.)
 //
@@ -698,12 +708,15 @@ struct pl_tex_params {
     // Arbitrary user data. libplacebo does not use this at all.
     void *user_data;
 
-    // Deprecated fields. These are now ignored entirely, and controlled
-    // via other mechanisms.
-    enum pl_tex_sample_mode sample_mode PL_DEPRECATED;
-    enum pl_tex_address_mode address_mode PL_DEPRECATED;
+    // Arbitrary identifying tag. Used only for debugging purposes.
+    pl_debug_tag debug_tag;
 };
 
+#define pl_tex_params(...) (&(struct pl_tex_params) {   \
+        .debug_tag = PL_DEBUG_TAG,                      \
+        __VA_ARGS__                                     \
+    })
+
 static inline int pl_tex_params_dimension(const struct pl_tex_params params)
 {
     return params.d ? 3 : params.h ? 2 : 1;
@@ -812,6 +825,8 @@ struct pl_tex_blit_params {
     enum pl_tex_sample_mode sample_mode;
 };
 
+#define pl_tex_blit_params(...) (&(struct pl_tex_blit_params) { __VA_ARGS__ })
+
 // Copy a sub-rectangle from one texture to another.
 void pl_tex_blit(pl_gpu gpu, const struct pl_tex_blit_params *params);
 
@@ -823,15 +838,23 @@ struct pl_tex_transfer_params {
     pl_tex tex;
 
     // Note: Superfluous parameters are ignored, i.e. for a 1D texture, the y
-    // and z fields of `rc`, as well as the corresponding strides, are ignored.
-    // In all other cases, the stride must be >= the corresponding dimension of
-    // `rc`, and the `rc` must be normalized and fully contained within the
-    // image dimensions. Missing fields in the `rc` are inferred from the image
-    // size. If unset, the strides are inferred from `rc` (that is, it's
-    // assumed that the data is tightly packed in the buffer).
+    // and z fields of `rc`, as well as the corresponding pitches, are ignored.
+    // In all other cases, the pitch must be large enough to contain the
+    // corresponding dimension of `rc`, and the `rc` must be normalized and
+    // fully contained within the image dimensions. Missing fields in the `rc`
+    // are inferred from the image size. If unset, the pitch is inferred
+    // from `rc` (that is, it's assumed that the data is tightly packed in the
+    // buffer). Otherwise, `row_pitch` *must* be a multiple of
+    // `tex->params.format->texel_align`, and `depth_pitch` must be a multiple
+    // of `row_pitch`.
     struct pl_rect3d rc;   // region of the texture to transfer
-    unsigned int stride_w; // the number of texels per horizontal row (x axis)
-    unsigned int stride_h; // the number of texels per vertical column (y axis)
+    size_t row_pitch;      // the number of bytes separating image rows
+    size_t depth_pitch;    // the number of bytes separating image planes
+
+    // Deprecated variants of `row_pitch` and `depth_pitch` for backwards
+    // compatibility with older versions of libplacebo. Avoid using.
+    unsigned int stride_w PL_DEPRECATED;
+    unsigned int stride_h PL_DEPRECATED;
 
     // An optional timer to report the approximate duration of the texture
     // transfer to. Note that this is only an approximation, since the actual
@@ -873,6 +896,8 @@ struct pl_tex_transfer_params {
     // texture format; i.e. there is no explicit conversion between formats.
 };
 
+#define pl_tex_transfer_params(...) (&(struct pl_tex_transfer_params) { __VA_ARGS__ })
+
 // Upload data to a texture. Returns whether successful.
 bool pl_tex_upload(pl_gpu gpu, const struct pl_tex_transfer_params *params);
 
@@ -1118,7 +1143,10 @@ struct pl_blend_params {
     enum pl_blend_mode dst_alpha;
 };
 
-extern const struct pl_blend_params pl_alpha_overlay; // typical alpha compositing
+#define pl_blend_params(...) (&(struct pl_blend_params) { __VA_ARGS__ })
+
+// Typical alpha compositing
+extern const struct pl_blend_params pl_alpha_overlay;
 
 enum pl_prim_type {
     PL_PRIM_TRIANGLE_LIST,
@@ -1126,6 +1154,12 @@ enum pl_prim_type {
     PL_PRIM_TYPE_COUNT,
 };
 
+enum pl_index_format {
+    PL_INDEX_UINT16 = 0,
+    PL_INDEX_UINT32,
+    PL_INDEX_FORMAT_COUNT,
+};
+
 enum pl_pass_type {
     PL_PASS_INVALID = 0,
     PL_PASS_RASTER,  // vertex+fragment shader
@@ -1167,10 +1201,14 @@ struct pl_pass_params {
     // the same shader. Can be used to speed up pass creation on already
     // known/cached shaders.
     //
-    // Note: There are no restrictions on this. Passing an out-of-date cache,
-    // passing a cache corresponding to a different program, or passing a cache
-    // belonging to a different GPU, are all valid. But obviously, in such cases,
-    // there is no benefit in doing so.
+    // Note: There are a few restrictions on this. Passing an out-of-date
+    // cache, passing a cache corresponding to a different program, or passing
+    // a cache belonging to a different GPU, are all guaranteed to be valid.
+    //
+    // It is, however, undefined behavior to pass arbitrary or maliciously
+    // crafted bytes - and users are advised that attaching a shader cache
+    // obtained from the internet could lead to arbitrary program behavior
+    // (possibly including code execution).
     const uint8_t *cached_program;
     size_t cached_program_len;
 
@@ -1180,21 +1218,18 @@ struct pl_pass_params {
     enum pl_prim_type vertex_type;
     struct pl_vertex_attrib *vertex_attribs;
     int num_vertex_attribs;
-    size_t vertex_stride;
+    size_t vertex_stride; // must be a multiple of limits.align_vertex_stride
 
     // The vertex shader itself.
     const char *vertex_shader;
 
-    // The target texture this render pass is intended to be used with. This
-    // doesn't have to come from a real texture - the caller can also invent
-    // values or pass a blank struct, as long as `target_dummy.params.format`
-    // is set. The format must support `PL_FMT_CAP_RENDERABLE`. If any other
-    // fields are set, the GPU may be able to further optimize the render pass
-    // for this particular type of texture.
-    PL_STRUCT(pl_tex) target_dummy;
+    // Target format. The format must support PL_FMT_CAP_RENDERABLE. The
+    // resulting pass may only be used on textures that have a format with a
+    // `pl_fmt.signature` compatible to this format.
+    pl_fmt target_format;
 
     // Target blending mode. If this is NULL, blending is disabled. Otherwise,
-    // the `target_dummy.params.format` must have PL_FMT_CAP_BLENDABLE.
+    // the `target_format` must also support PL_FMT_CAP_BLENDABLE.
     const struct pl_blend_params *blend_params;
 
     // If false, the target's existing contents will be discarded before the
@@ -1203,8 +1238,13 @@ struct pl_pass_params {
     //
     // Specifying `blend_params` requires `load_target` to be true.
     bool load_target;
+
+    // (Deprecated) Fallback for `target_format`.
+    PL_STRUCT(pl_tex) target_dummy PL_DEPRECATED;
 };
 
+#define pl_pass_params(...) (&(struct pl_pass_params) { __VA_ARGS__ })
+
 // Conflates the following typical GPU API concepts:
 // - various kinds of shaders
 // - rendering pipelines
@@ -1273,8 +1313,10 @@ struct pl_pass_run_params {
 
     // --- pass->params.type==PL_PASS_RASTER only
 
-    // Target must be a 2D texture, target->params.renderable must be true, and
-    // target->params.format must match pass->params.target_dummy.params.format.
+    // Target must be a 2D texture, `target->params.renderable` must be true,
+    // and `target->params.format->signature` must match the signature provided
+    // in `pass->params.target_format`.
+    //
     // If the viewport or scissors are left blank, they are inferred from
     // target->params.
     //
@@ -1296,11 +1338,12 @@ struct pl_pass_run_params {
     pl_buf vertex_buf;
     size_t buf_offset;
 
-    // (Optional) Index data may be provided in the form of `uint16_t` index
-    // offsets. These will be used for instanced rendering. Similar to
-    // vertex data, this can be provided in two forms:
+    // (Optional) Index data may be provided in the form given by `index_fmt`.
+    // These will be used for instanced rendering. Similar to vertex data, this
+    // can be provided in two forms:
     // 1. From host memory
-    const uint16_t *index_data;
+    const void *index_data;
+    enum pl_index_format index_fmt;
     // 2. From an index buffer (requires `index_buf->params.drawable`)
     pl_buf index_buf;
     size_t index_offset;
@@ -1315,6 +1358,8 @@ struct pl_pass_run_params {
     int compute_groups[3];
 };
 
+#define pl_pass_run_params(...) (&(struct pl_pass_run_params) { __VA_ARGS__ })
+
 // Execute a render pass.
 void pl_pass_run(pl_gpu gpu, const struct pl_pass_run_params *params);
 
